@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+from contextlib import contextmanager
+import errno
 import json
 from concurrent.futures import ProcessPoolExecutor
 import io
@@ -6,17 +8,11 @@ import re
 import time
 import pytest
 from itertools import chain
-from s3fs.core import S3FileSystem, FileNotFoundError
+from s3fs.core import S3FileSystem
 from s3fs.utils import seek_delimiter, ignoring, SSEParams
 import moto
 import boto3
-
-try:
-    from unittest import mock
-except ImportError:
-    # python 2.7
-    import mock
-
+from unittest import mock
 from botocore.exceptions import NoCredentialsError
 
 test_bucket_name = 'test'
@@ -55,9 +51,7 @@ d = test_bucket_name + '/tmp/test/d'
 @pytest.yield_fixture
 def s3():
     # writable local S3 system
-    try:
-        m = moto.mock_s3()
-        m.start()
+    with moto.mock_s3():
         import boto3
         client = boto3.client('s3')
         client.create_bucket(Bucket=test_bucket_name, ACL='public-read')
@@ -80,7 +74,8 @@ def s3():
                     "Effect": "Deny",
                     "Principal": "*",
                     "Action": "s3:PutObject",
-                    "Resource": "arn:aws:s3:::{bucket_name}/*".format(bucket_name=secure_bucket_name),
+                    "Resource": "arn:aws:s3:::{bucket_name}/*".format(
+                        bucket_name=secure_bucket_name),
                     "Condition": {
                         "StringNotEquals": {
                             "s3:x-amz-server-side-encryption": "aws:kms"
@@ -99,7 +94,9 @@ def s3():
         for flist in [files, csv_files, text_files, glob_files]:
             for f, data in flist.items():
                 client.put_object(Bucket=test_bucket_name, Key=f, Body=data)
-        yield S3FileSystem(anon=False)
+        s3 = S3FileSystem(anon=False)
+        s3.invalidate_cache()
+        yield s3
         for flist in [files, csv_files, text_files, glob_files]:
             for f, data in flist.items():
                 try:
@@ -115,8 +112,14 @@ def s3():
                 client.delete_object(Bucket=secure_bucket_name, Key=k)
             except:
                 pass
-    finally:
-        m.stop()
+
+
+@contextmanager
+def expect_errno(expected_errno):
+    """Expect an OSError and validate its errno code."""
+    with pytest.raises(OSError) as error:
+        yield
+    assert error.value.errno == expected_errno, 'OSError has wrong error code.'
 
 
 def test_simple(s3):
@@ -155,7 +158,7 @@ def test_tokenize():
 
 def test_idempotent_connect(s3):
     con1 = s3.connect()
-    con2 = s3.connect()
+    con2 = s3.connect(refresh=False)
     con3 = s3.connect(refresh=True)
     assert con1 is con2
     assert con1 is not con3
@@ -174,10 +177,10 @@ def test_info(s3):
     s3.touch(b)
     assert s3.info(a) == s3.ls(a, detail=True)[0]
     parent = a.rsplit('/', 1)[0]
-    s3.dirs.pop(a)  # remove full path from the cache
+    s3.invalidate_cache()  # remove full path from the cache
     s3.ls(parent)  # fill the cache with parent dir
-    assert s3.info(a) == s3.dirs[parent][0]  # correct value
-    assert id(s3.info(a)) == id(s3.dirs[parent][0])  # is object from cache
+    assert s3.info(a) == s3.dircache[parent][0]  # correct value
+    assert id(s3.info(a)) == id(s3.dircache[parent][0])  # is object from cache
 
     new_parent = test_bucket_name + '/foo'
     s3.mkdir(new_parent)
@@ -257,7 +260,7 @@ def test_not_delegate():
 def test_ls(s3):
     assert set(s3.ls('')) == {test_bucket_name,
                               secure_bucket_name, versioned_bucket_name}
-    with pytest.raises((OSError, IOError)):
+    with pytest.raises(FileNotFoundError):
         s3.ls('nonexistent')
     fn = test_bucket_name + '/test/accounts.1.json'
     assert fn in s3.ls(test_bucket_name + '/test')
@@ -324,20 +327,15 @@ def test_isdir(s3):
 
     assert not s3.isdir(b)
     assert not s3.isdir(b + '/')
-    s3.mkdir(b)
-    assert s3.isdir(b)
-    assert s3.isdir(b + '/')
 
     assert not s3.isdir(c)
     assert not s3.isdir(c + '/')
-    s3.mkdir(c + '/')
-    assert s3.isdir(c)
-    assert s3.isdir(c + '/')
 
     # test cache
-    assert not s3.dirs
+    s3.invalidate_cache()
+    assert not s3.dircache
     s3.ls(test_bucket_name + '/nested')
-    assert test_bucket_name + '/nested' in s3.dirs
+    assert test_bucket_name + '/nested' in s3.dircache
     assert not s3.isdir(test_bucket_name + '/nested/file1')
     assert not s3.isdir(test_bucket_name + '/nested/file2')
     assert s3.isdir(test_bucket_name + '/nested/nested2')
@@ -350,9 +348,9 @@ def test_rm(s3):
     assert s3.exists(a)
     s3.rm(a)
     assert not s3.exists(a)
-    with pytest.raises((OSError, IOError)):
+    with pytest.raises(FileNotFoundError):
         s3.rm(test_bucket_name + '/nonexistent')
-    with pytest.raises((OSError, IOError)):
+    with pytest.raises(FileNotFoundError):
         s3.rm('nonexistent')
     s3.rm(test_bucket_name + '/nested', recursive=True)
     assert not s3.exists(test_bucket_name + '/nested/nested2/file1')
@@ -364,19 +362,6 @@ def test_rm(s3):
 
 
 def test_rmdir(s3):
-    nested = a + "/nested"
-    s3.touch(a)
-    s3.touch(nested)
-    s3.touch(b)
-
-    with pytest.raises(IOError):  # doesn't delete non empty directory
-        s3.rmdir(a)
-
-    s3.rmdir(nested)
-    s3.rmdir(a)
-    assert a not in s3.ls(test_bucket_name)
-    assert nested not in s3.ls(test_bucket_name)
-
     bucket = 'test1_bucket'
     s3.mkdir(bucket)
     s3.rmdir(bucket)
@@ -409,11 +394,11 @@ def test_mkdir_client_region_name():
 
 
 def test_bulk_delete(s3):
-    with pytest.raises((OSError, IOError)):
+    with pytest.raises(FileNotFoundError):
         s3.bulk_delete(['nonexistent/file'])
     with pytest.raises(ValueError):
         s3.bulk_delete(['bucket1/file', 'bucket2/file'])
-    filelist = s3.walk(test_bucket_name + '/nested')
+    filelist = s3.find(test_bucket_name+'/nested')
     s3.bulk_delete(filelist)
     assert not s3.exists(test_bucket_name + '/nested/nested2/file1')
 
@@ -423,7 +408,8 @@ def test_anonymous_access():
         s3 = S3FileSystem(anon=True)
         assert s3.ls('') == []
         # TODO: public bucket doesn't work through moto
-    with pytest.raises((OSError, IOError)):
+
+    with pytest.raises(PermissionError):
         s3.mkdir('newbucket')
 
 
@@ -439,11 +425,11 @@ def test_s3_file_access(s3):
 def test_s3_file_info(s3):
     fn = test_bucket_name + '/nested/file1'
     data = b'hello\n'
-    assert fn in s3.walk(test_bucket_name)
+    assert fn in s3.find(test_bucket_name)
     assert s3.exists(fn)
     assert not s3.exists(fn + 'another')
     assert s3.info(fn)['Size'] == len(data)
-    with pytest.raises((OSError, IOError)):
+    with pytest.raises(FileNotFoundError):
         s3.info(fn + 'another')
 
 
@@ -456,7 +442,7 @@ def test_bucket_exists(s3):
 
 
 def test_du(s3):
-    d = s3.du(test_bucket_name, deep=True)
+    d = s3.du(test_bucket_name, total=False)
     assert all(isinstance(v, int) and v >= 0 for v in d.values())
     assert test_bucket_name + '/nested/file1' in d
 
@@ -477,9 +463,9 @@ def test_s3_ls(s3):
 def test_s3_big_ls(s3):
     for x in range(1200):
         s3.touch(test_bucket_name + '/thousand/%i.part' % x)
-    assert len(s3.walk(test_bucket_name)) > 1200
+    assert len(s3.find(test_bucket_name)) > 1200
     s3.rm(test_bucket_name + '/thousand/', recursive=True)
-    assert len(s3.walk(test_bucket_name + '/thousand/')) == 0
+    assert len(s3.find(test_bucket_name + '/thousand/')) == 0
 
 
 def test_s3_ls_detail(s3):
@@ -496,7 +482,7 @@ def test_s3_glob(s3):
     assert fn in s3.glob(test_bucket_name + '/nested/file*')
     assert fn in s3.glob(test_bucket_name + '/*/*')
     assert all(any(p.startswith(f + '/') or p == f
-                   for p in s3.walk(test_bucket_name, directories=True))
+                   for p in s3.find(test_bucket_name))
                for f in s3.glob(test_bucket_name + '/nested/*'))
     assert [test_bucket_name +
             '/nested/nested2'] == s3.glob(test_bucket_name + '/nested/nested2')
@@ -573,7 +559,7 @@ def test_seek(s3):
 
 
 def test_bad_open(s3):
-    with pytest.raises(IOError):
+    with pytest.raises(ValueError):
         s3.open('')
 
 
@@ -599,28 +585,28 @@ def test_get_put(s3, tmpdir):
     assert open(test_file, 'rb').read() == data
     s3.put(test_file, test_bucket_name + '/temp')
     assert s3.du(test_bucket_name +
-                 '/temp')[test_bucket_name + '/temp'] == len(data)
+                 '/temp', total=False)[test_bucket_name + '/temp'] == len(data)
     assert s3.cat(test_bucket_name + '/temp') == data
 
 
 def test_errors(s3):
-    with pytest.raises((IOError, OSError)):
+    with pytest.raises(FileNotFoundError):
         s3.open(test_bucket_name + '/tmp/test/shfoshf', 'rb')
 
     # This is fine, no need for interleaving directories on S3
     # with pytest.raises((IOError, OSError)):
     #    s3.touch('tmp/test/shfoshf/x')
 
-    with pytest.raises((IOError, OSError)):
+    with pytest.raises(FileNotFoundError):
         s3.rm(test_bucket_name + '/tmp/test/shfoshf/x')
 
-    with pytest.raises((IOError, OSError)):
+    with pytest.raises(FileNotFoundError):
         s3.mv(test_bucket_name + '/tmp/test/shfoshf/x', 'tmp/test/shfoshf/y')
 
-    with pytest.raises((IOError, OSError)):
+    with pytest.raises(ValueError):
         s3.open('x', 'rb')
 
-    with pytest.raises(IOError):
+    with pytest.raises(FileNotFoundError):
         s3.rm('unknown')
 
     with pytest.raises(ValueError):
@@ -632,14 +618,14 @@ def test_errors(s3):
         f.close()
         f.read()
 
-    with pytest.raises((IOError, OSError)):
+    with pytest.raises(ValueError):
         s3.mkdir('/')
 
     with pytest.raises(ValueError):
-        s3.walk('')
+        s3.find('')
 
     with pytest.raises(ValueError):
-        s3.walk('s3://')
+        s3.find('s3://')
 
 
 def test_read_small(s3):
@@ -694,15 +680,14 @@ def test_new_bucket(s3):
     assert s3.exists('new')
     with s3.open('new/temp', 'wb') as f:
         f.write(b'hello')
-    with pytest.raises((IOError, OSError)):
+    with expect_errno(errno.ENOTEMPTY):
         s3.rmdir('new')
-    with pytest.raises((IOError, OSError)):
-        s3.rmdir('new/temp')
+
     s3.rm('new/temp')
     s3.rmdir('new')
     assert 'new' not in s3.ls('')
     assert not s3.exists('new')
-    with pytest.raises((IOError, OSError)):
+    with pytest.raises(FileNotFoundError):
         s3.ls('new')
 
 
@@ -803,7 +788,7 @@ def test_write_fails(s3):
     f.close()
     with pytest.raises(ValueError):
         f.write(b'hello')
-    with pytest.raises((OSError, IOError)):
+    with pytest.raises(FileNotFoundError):
         s3.open('nonexistentbucket/temp', 'wb').close()
 
 
@@ -838,43 +823,6 @@ def test_readline(s3):
             assert result == expected
 
 
-def test_readline_from_cache(s3):
-    data = b'a,b\n11,22\n3,4'
-    with s3.open(a, 'wb') as f:
-        f.write(data)
-
-    with s3.open(a, 'rb') as f:
-        result = f.readline()
-        assert result == b'a,b\n'
-        assert f.loc == 4
-        assert f.cache == data
-
-        result = f.readline()
-        assert result == b'11,22\n'
-        assert f.loc == 10
-        assert f.cache == data
-
-        result = f.readline()
-        assert result == b'3,4'
-        assert f.loc == 13
-        assert f.cache == data
-
-
-def test_readline_partial(s3):
-    data = b'aaaaa,bbbbb\n12345,6789\n'
-    with s3.open(a, 'wb') as f:
-        f.write(data)
-    with s3.open(a, 'rb') as f:
-        result = f.readline(5)
-        assert result == b'aaaaa'
-        result = f.readline(5)
-        assert result == b',bbbb'
-        result = f.readline(5)
-        assert result == b'b\n'
-        result = f.readline()
-        assert result == b'12345,6789\n'
-
-
 def test_readline_empty(s3):
     data = b''
     with s3.open(a, 'wb') as f:
@@ -921,8 +869,6 @@ def test_iterable(s3):
         assert f.readline() == b'123'
         f.seek(1)
         assert f.readline() == b'bc\n'
-        assert f.readline(1) == b'1'
-        assert f.readline() == b'23'
 
     with s3.open(a) as f:
         out = list(f)
@@ -982,6 +928,8 @@ def test_append(s3):
     assert s3.cat(a) == b'a' * 10 * 2 ** 20
 
     with s3.open(a, 'ab') as f:
+        assert f.parts is None
+        f._initiate_upload()
         assert f.parts
         assert f.tell() == 10 * 2 ** 20
         f.write(b'extra')  # append, small write, big file
@@ -1006,6 +954,9 @@ def test_bigger_than_block_read(s3):
 
 
 def test_current(s3):
+    s3._cache.clear()
+    s3 = S3FileSystem()
+    assert s3.current() is s3
     assert S3FileSystem.current() is s3
 
 
@@ -1022,7 +973,7 @@ def test_array(s3):
 
 
 def _get_s3_id(s3):
-    return id(s3.connect())
+    return id(s3.s3)
 
 
 def test_no_connection_sharing_among_processes(s3):
@@ -1045,13 +996,13 @@ def test_public_file(s3):
         s3.touch(other_bucket_name + '/afile', acl='public-read')
 
         s = S3FileSystem(anon=True)
-        with pytest.raises((IOError, OSError)):
+        with pytest.raises(PermissionError):
             s.ls(test_bucket_name)
         s.ls(other_bucket_name)
 
         s3.chmod(test_bucket_name, acl='public-read')
         s3.chmod(other_bucket_name, acl='private')
-        with pytest.raises((IOError, OSError)):
+        with pytest.raises(PermissionError):
             s.ls(other_bucket_name, refresh=True)
         assert s.ls(test_bucket_name, refresh=True)
 
@@ -1086,25 +1037,6 @@ def test_multipart_upload_blocksize(s3):
     # Ensure that the multipart upload consists of only 3 parts
     assert len(s3f.parts) == expected_parts
     s3f.close()
-    assert len(s3f.parts) == 0
-
-
-def test_not_fill_cache(s3):
-    fname = list(files)[0]
-    data = files[fname]
-    with s3.open(test_bucket_name + '/' + fname, 'rb', block_size=5,
-                 fill_cache=False) as f:
-        f.read(6)
-        assert len(f.cache) == 11
-        assert data.startswith(f.cache)
-        f.seek(10)
-        f.read(2)
-        assert f.start < 10  # keeps some old data
-
-        f.seek(31)
-        out = f.read(6)
-        assert len(f.cache) == 11  # dropped
-        assert out == data[31:37]
 
 
 def test_default_pars(s3):
@@ -1229,30 +1161,35 @@ def test_readinto(s3):
     assert contents.startswith(b'Hello, World!')
 
 
-def test_change_defaults_only_subsequent(monkeypatch):
+def test_change_defaults_only_subsequent():
     """Test for Issue #135
 
     Ensure that changing the default block size doesn't affect existing file
     systems that were created using that default. It should only affect file
     systems created after the change.
     """
-    fs_default = S3FileSystem()
-    assert fs_default.default_block_size == 5 * (1024 ** 2)
+    try:
+        S3FileSystem.cachable = False  # don't reuse instances with same pars
 
-    fs_overridden = S3FileSystem(default_block_size=64 * (1024 ** 2))
-    assert fs_overridden.default_block_size == 64 * (1024 ** 2)
+        fs_default = S3FileSystem()
+        assert fs_default.default_block_size == 5 * (1024 ** 2)
 
-    # Suppose I want all subsequent file systems to have a block size of 1 GiB
-    # instead of 5 MiB:
-    monkeypatch.setattr(S3FileSystem, 'default_block_size', 1024 ** 3)
+        fs_overridden = S3FileSystem(default_block_size=64 * (1024 ** 2))
+        assert fs_overridden.default_block_size == 64 * (1024 ** 2)
 
-    fs_big = S3FileSystem()
-    assert fs_big.default_block_size == 1024 ** 3
+        # Suppose I want all subsequent file systems to have a block size of 1 GiB
+        # instead of 5 MiB:
+        S3FileSystem.default_block_size = 1024 ** 3
 
-    # Test the other file systems created to see if their block sizes changed
-    assert fs_overridden.default_block_size == 64 * (1024 ** 2)
-    assert fs_default.default_block_size == 5 * (1024 ** 2)
+        fs_big = S3FileSystem()
+        assert fs_big.default_block_size == 1024 ** 3
 
+        # Test the other file systems created to see if their block sizes changed
+        assert fs_overridden.default_block_size == 64 * (1024 ** 2)
+        assert fs_default.default_block_size == 5 * (1024 ** 2)
+    finally:
+        S3FileSystem.default_block_size = 5 * (1024 ** 2)
+        S3FileSystem.cachable = True
 
 def test_passed_in_session_set_correctly(s3):
     session = boto3.session.Session()
